/* MACE_GO: MAximal Clique Enumerater - Greater size/Optimum */
/* ver 1.0 1/Sep/2005 Takeaki Uno   e-mail:uno@nii.jp, 
    homepage:   http://research.nii.ac.jp/~uno/index.html  */
/* This program is available for only academic use, basically.
   Anyone can modify this program, but he/she has to write down 
    the change of the modification on the top of the source code.
   Neither contact nor appointment to Takeaki Uno is needed.
   If one wants to re-distribute this code, do not forget to 
    refer the newest code, and show the link to homepage of 
    Takeaki Uno, to notify the news about this code for the users.
   For the commercial use, please make a contact to Takeaki Uno. */

//graphs:
//Linkedln 2006.8
//U.S. Patent at 1985
// YahooWeb

#ifndef _mace_go_c_
#define _mace_go_c_

#define WEIGHT_DOUBLE

#include"sgraph.c"
#include"problem.c"


//int **PP->shift;    /* pointers to the positions of the current processing items in each transaction */


void MACEGO_error (){
  ERROR_MES = "command explanation";
  print_err ("mace_go MCqVe [options] input-filename [output-filename]\n\
%%:show progress, _:no message, +:write solutions in append mode\n\
C:enumerate cliques, M:find maximum clique, e:edge_list format for input file\n\
[options]\n\
-M [num]:enumerate maximal cliques of size at least [num]\n\
-# [num]:stop after outputting [num] solutions\n\
-, [char]:give the separator of the numbers in the output\n\
-Q [filename]:replace the output numbers according to the permutation table given by [filename]\n\
# the 1st letter of input-filename cannot be '-'.n\
# if the output file name is -, the solutions will be output to standard output.\n");
  EXIT;
}


/***************************************************/
/*  read parameters from command line              */
/***************************************************/
void MACEGO_read_param (int argc, char *argv[], PROBLEM *PP){
  ITEMSET *II=&PP->II;
  int c=1;
  if ( argc < c+2 ){ MACEGO_error (); return; }
  II->lb = 2;

  if ( !strchr (argv[c], '_') ){ II->flag |= SHOW_MESSAGE; PP->FF.flag |= SHOW_MESSAGE; }
  if ( strchr (argv[c], '%') ) II->flag |= SHOW_PROGRESS;
  if ( strchr (argv[c], '+') ) II->flag |= ITEMSET_APPEND;
  if ( strchr (argv[c], 'M') ){ PP->problem = PROBLEM_MAXIMAL; II->frq = 0;}
  else if ( strchr (argv[c], 'C') ) PP->problem = PROBLEM_FREQSET;
  else error ("C command has to be specified", EXIT);
  if ( strchr (argv[c], 'V') ) II->flag |= SHOW_PROGRESS;
  if ( strchr (argv[c], 'e') ) PP->SG.flag |= LOAD_ELE;
  c++;

  while ( argv[c][0] == '-' ){
    switch (argv[c][1]){
      case 'M': II->lb = atoi(argv[c+1]);  // lower bound for size
      break; case 'm': II->frq = atof(argv[c+1]);   // lower bound for weight
      break; case 'w': PP->SG.rwfname = argv[c+1];
      break; case '#': II->max_solutions = atoi(argv[c+1]);
      break; case ',': II->separator = argv[c+1][0];
      break; case 'Q': PP->outperm_fname = argv[c+1];
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+1 ){ MACEGO_error (); return; }
  }

  NEXT:;
  PP->SG.fname = argv[c];
  if ( argc>c+1 ) PP->output_fname = argv[c+1];
}



/* make the subgraph induced by edges which has at least one endpoint in the neighbor of v,
  the edges incident to v will be removed. */
/* the node number is not preserved, but the inverse permutation is written at edge.v[i].end */
/* return, pivot vertex of the new graph */
QUEUE_INT MACEGO_mk_induced_graph (SGRAPH *GG, SGRAPH *G, QUEUE_INT v, QUEUE_INT *occ_t, QUEUE_INT pivot, QUEUE_INT last, QUEUE_INT *new_p){
  QUEUE_INT m=0, p=0, c, *x, *y=G->edge.v[pivot].v, i=1, ii=G->edge.v[v].t, j;
  MQUE_FLOOP (G->edge.v[v], x){
    while (*y < *x) y++;
    G->edge.v[*x].s = (*x<last&&(*x==*y||*x<v))? i++: ii--;
  }
  if ( ii==0 ){
    MQUE_FLOOP (G->edge.v[v], x) G->edge.v[*x].s = 0; 
    return (0);
  }

  MQUE_FLOOP (G->edge.v[v], x){
    i = G->edge.v[*x].s - 1;
    if ( i >= ii ) continue;
    c=0;MQUE_FLOOP (G->edge.v[*x], y){
      if ( G->edge.v[*y].s == 0 ) continue;
      occ_t[i]++;
      if ( G->edge.v[*y].s > ii ) occ_t[G->edge.v[*y].s-1]++;
      else c++;
    }
    if ( c>m ){ m=c; p=i; }
  }
//  MQUE_FLOOP (G->edge.v[v], x) printf ("(%d,%d) ", *x, G->edge.v[*x].s);
//  printf ("::::: %d ii\n", ii);
  MQUE_ALLOC (GG->edge.v, G->edge.v[v].t, occ_t, 0, 2, EXIT);
  GG->edge.t = G->edge.v[v].t;
  GG->edge.buf = GG->edge.v[0].v; // pointer to the edge buffer, to be freed
  FLOOP (i, 0, GG->edge.t) GG->edge.v[i].end = occ_t[i];

  MQUE_FLOOP (G->edge.v[v], x){
    i = G->edge.v[*x].s - 1;
    if ( i >= ii ) continue;
    MQUE_FLOOP (G->edge.v[*x], y){
      j = G->edge.v[*y].s;
      if ( j == 0 ) continue;
      j--;
      if ( j >= ii ){
        GG->edge.v[i].v[--GG->edge.v[i].end] = j;
        QUE_INS (GG->edge.v[j], i);
      } else QUE_INS (GG->edge.v[i], j);
    }
  }
  FLOOP (i, 0, GG->edge.t){
    GG->edge.v[i].t = occ_t[i]; // recover the queue size
    GG->edge.v[i].v[GG->edge.v[i].t] = GG->edge.t+1; // put endmark
    occ_t[i] = 0;
  }
  MQUE_FLOOP (G->edge.v[v], x){
    GG->edge.v[G->edge.v[*x].s-1].end = G->edge.v[*x].end; // inv. perm
    G->edge.v[*x].s = 0; 
  }
  *new_p = p; return (ii);
}

/* remove a node from a graph with decreasing the degree of other vertices but not their incidnet edges */
void MACEGO_rm_node (SGRAPH *G, QUEUE_INT v, QUEUE_INT d){
  QUEUE_INT *x, *xx, *xi=G->edge.v[v].v;

  MQUE_FLOOP2 (G->edge.v[v], x, xx){      // remove edges
    if ( G->edge.v[*x].t == 0 ){ xx++; continue; } // skip removed operated small degree node
      // decrease the degree without writting delete mark
    if ( --G->edge.v[*x].t < d ){ *xi = *x; xi++; }
  }
  G->edge.v[v].t = 0;
//  for (xi=x=G->edge.v[v].v,xx=x+t ; x<xx ; x++){
    FLOOP (x, G->edge.v[v].v, xi){
    if ( G->edge.v[*x].t > 0 ) MACEGO_rm_node (G, *x, d); // recursively remove node
  }
}

/* remove the nodes with degrees smaller than d (consider/update only nodes in occ) */
/* the maximum size of cand has to be twice #nodes */
/* assume that any node in cand is connected only to nodes in cand */
/* does not change the order of cand */
void MACEGO_rm_small_deg (SGRAPH *G, QUEUE_INT d){
  QUEUE_INT i, *x, *xx, *xi;
  
  FLOOP (i, 0, G->edge.t){
    if ( G->edge.v[i].t == 0 ) continue;
    if ( G->edge.v[i].t < d ) MACEGO_rm_node (G, i, d); // recursively remove node
  }
  FLOOP (i, 0, G->edge.t){
    if ( G->edge.v[i].t < d ) continue;  // skip small degree node
    xi=G->edge.v[i].v; MQUE_FLOOP2 (G->edge.v[i], x, xx){
      if ( G->edge.v[*x].t < d ){ xx++; continue; } // skip removed operated small degree node
        // in the case that the edge remains
      if ( x != xi ) *xi = *x;
      xi++;
    }
    *xi = G->edge.t;  // endmark
    G->edge.v[i].t = xi - G->edge.v[i].v;
  }
}

/* remove edge (u,v) from the edge list of u */
void MACEGO_remove_edge (SGRAPH *G, QUEUE_INT v, QUEUE_INT *u, QUEUE *cand){
  if ( G->edge.v[*u].s == G->edge.t ){
    QUE_INS (*cand, *u); G->edge.v[*u].s = 0;  // non-changed node is activated
  }
  while (G->edge.v[*u].v[G->edge.v[*u].s] < v) G->edge.v[*u].s++;
  G->edge.v[*u].v[G->edge.v[*u].s] = G->edge.t;   // write delete mark!
  G->edge.v[*u].t--;
  G->edge.v[*u].s++;
  *u = G->edge.t; // write delete mark!
  G->edge.v[v].t--;
}

/* remove all edges adjacent to v*/
void MACEGO_remove_all_edges (SGRAPH *G, QUEUE_INT v, QUEUE *cand, int d){
  QUEUE_INT *x, *xx, *xi=G->edge.v[v].v;

  MQUE_FLOOP2 (G->edge.v[v], x, xx){      // remove edges
    if ( *x >= G->edge.t ){ xx++; continue; } // skip removed edge, and increase the end
    if ( G->edge.v[*x].t == 0 ){ xx++; continue; } // skip removed operated small degree node
    if ( G->edge.v[*x].s == G->edge.t ){
      QUE_INS (*cand, *x); G->edge.v[*x].s = 0;  // non-changed node is activated
    }
    G->edge.v[*x].t--;  // decrease the degree without writting delete mark
    if ( G->edge.v[*x].t < d ){ *xi = *x; xi++; }
  }
  G->edge.v[v].t = 0;
  FLOOP (x, G->edge.v[v].v, xi) MACEGO_remove_all_edges (G, *x, cand, d);
  G->edge.v[v].v[0] = G->edge.t+1;  // endmark
}

/* remove edges (u,v) incident to a node v s.t. u and v have common neighbors less than d */
/* assume occ_w's are zero for nessesary positions
   assume jump.s =0, assume edge.s = 0, and cand is in increasing order */
/* G->edge.v[*x].s  is the shift-counter */
/* assume end mark of no less than G->edge.t at the end of each edge list */
void MACEGO_rm_non_common_edge (SGRAPH *G, int d, QUEUE *cand2, WEIGHT *occ_w, QUEUE *jump){
  QUEUE_INT *v, *vv, *x, *xx, *xi, *y, i;
  int flag;

  cand2->t = 0;
    // set shift-counter to un-initialized  
  FLOOP (i, 0, G->edge.t)
      if ( G->edge.v[i].t >= d ) QUE_INS (*cand2, i);

  BEGIN:;
  MQUE_FLOOP2 (*cand2, v, vv){
    flag = 0;
      // delivery
    jump->t = 0;
    MQUE_FLOOP2 (G->edge.v[*v], x, xx){
      if ( *x >= G->edge.t ){ xx++; continue; } // skip removed edge, and increase the end
      if ( G->edge.v[*x].t == 0 ){ xx++; continue; } // skip removed operated small degree node
      MQUE_MLOOP (G->edge.v[*x], y, *v){
        if ( occ_w[*y] == 0 ) QUE_INS (*jump, *y);
        occ_w[*y] += 1;
      }
    }
      // remove edges
    MQUE_FLOOP2 (G->edge.v[*v], x, xx){
      if ( *x >= G->edge.t ){ xx++; continue; } // skip removed edge
      if ( *x >= *v ) break; // remove only for edges to smaller index nodes
      if ( G->edge.v[*x].t == 0 ){ xx++; continue; } // skip removed operated small degree node
      if ( occ_w[*x] < d-1 ){
        if ( G->edge.v[*v].t == d ){
          MACEGO_remove_all_edges (G, *v, cand2, d);
          break;
        } else if ( G->edge.v[*x].t == d ){
          MACEGO_remove_all_edges (G, *x, cand2, d);
        } else MACEGO_remove_edge (G, *v, x, cand2);
      }
    }
      // clear counters
    MQUE_FLOOP (*jump, x) occ_w[*x] = 0;
  }
  if ( cand2->t == 0 ){
    FLOOP (i, 0, G->edge.t) G->edge.v[i].s = 0;
    return;
  }

  qsort_QUEUE_INT (cand2->v, cand2->t, 1);  // sort the candidates because it has recovered nodes on the tail

    // remove the edges with the delete mark, and connected to small degree node
  vv=cand2->v; MQUE_FLOOP (*cand2, v){
    if ( G->edge.v[*v].t < d ) continue;  // skip small degree node
    xi=G->edge.v[*v].v; MQUE_FLOOP2 (G->edge.v[*v], x, xx){
      if ( *x >= G->edge.t ){ xx++; continue; } // skip removed edge
      if ( G->edge.v[*x].t == 0 ){ xx++; continue; } // skip removed operated small degree node
        // in the case that the edge remains
      if ( x != xi ) *xi = *x;
      xi++;
    }
    if ( *xi == G->edge.t+1 ){ G->edge.v[*v].s = G->edge.t; continue; } // remove non-changed node from the queue, and put "removed" mark
    *xi = G->edge.t+1;  // endmark
    if ( v != vv ) *vv = *v;
    vv++;
    G->edge.v[*v].s = 0;
  }
  cand2->t = vv - cand2->v;
  goto BEGIN;
}

/******************************************************************/
/* iteration of clique enumeration   */
/******************************************************************/
void MACEGO_iter (PROBLEM *PP, SGRAPH *G, QUEUE_INT pivot, QUEUE_INT last){
  SGRAPH GG=INIT_SGRAPH;
  ITEMSET *II=&PP->II;
  QUEUE_INT ii, *y, i, p;
  char *ad;
  
  II->iters++;
//printf ("pivot=%d, last=%d \n", pivot, last);
    // reduce the graph by constraint of #common neighbors
  if ( II->lb >= 2 ) 
      MACEGO_rm_non_common_edge (G, II->lb, &PP->itemcand, PP->occ_w, &PP->itemjump);

  calloc2 (ad, G->edge.t, EXIT);
  MQUE_FLOOP (G->edge.v[pivot], y) ad[*y] = 1;
  y = G->edge.v[pivot].v;
  
  FLOOP (i, 0, last){
    if ( i == *y ){ y++; continue; }  // covered by pivot vertex
//printf ("----- %d %d: t=%d < %d\n", i, G->edge.v[i].end, G->edge.v[i].t, II->lb);
//QUEUE_print__ (&II->itemset);
    if ( G->edge.v[i].t < II->lb ) continue;
    if (  G->edge.v[i].t > 0 && G->edge.v[i].v[0] > i && ad[*y]==0 ) continue; // not a maximal
    QUE_INS (II->itemset, G->edge.v[i].end);
//printf ("i=%d, .end=%d  <%d\n", i, G->edge.v[i].t, last);
//QUEUE_print__ (&II->itemset);
//printf ("================== y=%d, %d   ", *y, pivot); QUEUE_print__ (&G->edge.v[pivot]);
    if ( G->edge.v[i].t == 0 ) ITEMSET_output_itemset (II, NULL, 0);
    else {
      ii = MACEGO_mk_induced_graph (&GG, G, i, PP->occ_t, pivot, last, &p);
      if ( ii>0 ){
//SGRAPH_print (G);
//QUEUE_print__ (&II->itemset);
//SGRAPH_print (&GG);
//QUEUE_INT ii;
//FLOOP (ii, 0, GG.edge.t) printf ("%d ", GG.edge.v[ii].end); printf ("\n");

        II->lb--;
        MACEGO_iter (PP, &GG, p, ii);
        SGRAPH_end (&GG);
        II->lb++;
      }
    }
    II->itemset.t--;
  }
  free (ad);
}

/***************************************************/
/*  main routine                                   */
/***************************************************/
int MACE_main (int argc, char *argv[]){
  PROBLEM PP;
  SGRAPH *G=&PP.SG;
  QUEUE_INT m, i, pivot;
  
  PROBLEM_init (&PP);
  MACEGO_read_param (argc, argv, &PP);
if ( ERROR_MES ) return (1);

  G->flag |= LOAD_INCSORT + LOAD_RM_DUP + LOAD_EDGE;
  PROBLEM_load (&PP);  if (ERROR_MES) return(1);
  if (G->rwfname) SETFAMILY_load_row_weight (&G->edge);

  FLOOP (i, 0, G->edge.t) G->edge.v[i].v[G->edge.v[i].t] = G->edge.t+1; // put endmark
  PROBLEM_alloc (&PP, G->edge.t, G->edge.t, G->edge.eles, NULL, PROBLEM_ITEMJUMP + PROBLEM_ITEMCAND +PROBLEM_OCC_T + PROBLEM_OCC_W);  if (ERROR_MES) return(1);
  FLOOP (i, 0, G->edge.t) G->edge.v[i].end = i;
  PP.dir = PP.II.lb = PP.II.lb-1;
  MACEGO_rm_small_deg (G, PP.II.lb);
//  SGRAPH_sort_node (G, 1);
  ST_MAX (m, pivot, G->edge.v, t, 0, G->edge.t);
  
//SGRAPH_print (G);
  MACEGO_iter (&PP, G, pivot, G->edge.t);
  ITEMSET_last_output (&PP.II);
  
  PROBLEM_end (&PP);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (MACE_main (argc, argv) );
}
#endif
/*******************************************************************************/


#endif

